/**********************************************************************
 * squishy_bugs.c                                          October 1999
 * Horms                                            horms@vergenet.net.
 * Verge Systems International                    http://vergenet.net./
 *
 * Squishy Bugs                                      
 * Squish the bugs with your mouse as they crawl up your terminal.
 * 
 * Written for Felicity <flea@flea.org>
 * Code hacked from slhelp.c, as distributed with slang
 * Released under the terms of the GNU GPL
 **********************************************************************/

#include "squishy_bugs.h"

int main (int args, char **argv){
  char *bugs;
  int cells;
  int population;
  int escaped;
  int squished;
  int rnd;
  int x;
  int y;
  int b;
  char *string;
  char *squish_string;
  int stay;

  extern int SLtt_Screen_Rows;
  extern int SLtt_Screen_Cols;

  if(SLang_init_tty (7, 1, 1)<0){
    fprintf(stderr, "main: SLang_init_tty");
    goto leave;
  }
  SLtt_get_terminfo ();
  SLtt_init_video ();
  SLtt_Use_Ansi_Colors = 0;

#ifdef SIGWINCH
  signal (SIGWINCH, SBtt_get_screen_size);
#endif

  SBtt_get_screen_size (0);
  SLtt_set_mouse_mode(1, 0);
  SBtt_cls ();

  cells=(SLtt_Screen_Rows-1)*SLtt_Screen_Cols;
  if((bugs=(char *)malloc(sizeof(char)*cells))==NULL){
    goto leave;
  }

  if((string=(char *)malloc(sizeof(char)*1024))==NULL){
    goto leave;
  }
  
  memset(bugs, NOBUG, cells);

  srand(time(NULL));
  population=0;
  escaped=0;
  squished=0;
  x=0;
  y=0;
  b=0;
  stay=0;
  squish_string="";
  while(1){
    if(
      (27 == SB_get_key(0))
      && ('[' == SB_get_key(0))
      && ('M' == SB_get_key(0))
    ){
      b = SB_get_key(0) - ' ';
      x = SB_get_key(0) - ' ' -1;
      y = SB_get_key(0) - ' ' -1;
      rnd=(rnd=SB_kill_bug_mouse(&bugs, x, y, b, 1, cells)>0)?rnd:0;
      if(rnd>0){
        population-=rnd;
        squished+=rnd;
        stay=SQUISH_STRING_STAY;
        squish_string=SQUISH_STRING;
      }
    }
    

    population+=SB_new_bug(&bugs, population, 20, cells);

    if(population>0){
      rnd=rand()%cells;
      while(*(bugs+rnd)!=BUG){
        rnd=(rnd+1)%cells;
      }
      rnd=(rnd=SB_move_bug(&bugs, rnd)>0)?rnd:0;
      population-=rnd;
      escaped+=rnd;
    }

    SLtt_goto_rc(SLtt_Screen_Rows-1, 0);
    sprintf(
      string, 
      "population: %04d escaped: %04d squished: %04d score: %05d %10s",
      population,
      escaped,
      squished,
      squished-escaped,
      squish_string
    );
    SLtt_write_string(string);
    SLtt_flush_output();
    usleep(250000/(population+1));
    if(stay--<1){
      squish_string="";
    }
  }
     


  leave:
  SLtt_reset_video ();
  SLang_reset_tty ();

  return(0);
}


/**********************************************************************
 * SBtt_get_screen_size
 * Get the screen size, can be used as a signal handler for SIGWINCH
 * if it exists on the system
 *
 * Borrowed from slhelp.c
 **********************************************************************/

static void SBtt_get_screen_size (int sig) {
   SLtt_get_screen_size ();
#ifdef SIGWINCH
   signal (SIGWINCH, get_screen_size);
#endif
}


/**********************************************************************
 * SBtt_cls
 * Clear the screen
 * 
 * Borrowed from slhelp.c
 **********************************************************************/

static void SBtt_cls (void) {
   SLtt_cls ();
   SLtt_flush_output ();
}


/**********************************************************************
 * SB_new_bug
 * Return: 0 if no new bug is created
 *         1 if a new bug is created
 **********************************************************************/

int SB_new_bug(char **bugs, int population, int density, int cells){
  int rnd;
  int rnd1;
  int current_density;
  int skipped;

  extern int SLtt_Screen_Cols;

  current_density=population/cells;
  if(current_density>=density){
    return(0);
  }

  if(population==0 || rand()%(density-current_density)==0){
    rnd1=rnd=rand()%SLtt_Screen_Cols;
    skipped=0;
    while(*((*bugs)+(cells-rnd-1))==BUG){
      rnd=(rnd+1)%SLtt_Screen_Cols;
      skipped++;
      if(rnd==rnd1||skipped>3){
        return(0);
      }
    }
    SB_put_bug_cell(bugs, BUG, cells-rnd-1);
    return(1);
  }

  return(0);
}


/**********************************************************************
 * SB_move_bug
 * Move a bug up the screen
 * Return: -1 if there is no bug in cell
 *         -2 if there is already a bug in the destination cell
 *         1  if the bug is moved off the screen
 *         0  if the bug is moved on the screen
 **********************************************************************/

int SB_move_bug(char **bugs, int cell){
  extern int SLtt_Screen_Cols;

  int newcell;

  if(*((*bugs)+cell)!=BUG){
    return(-1);
  }

  newcell=cell-SLtt_Screen_Cols;
  if(newcell<0){
    SB_put_bug_cell(bugs, NOBUG, cell);
    return(1);
  }

  if(*((*bugs)+newcell)==BUG){
    return(SB_move_bug(bugs, newcell));
  }
  SB_put_bug_cell(bugs, NOBUG, cell);
  SB_put_bug_cell(bugs, BUG, newcell);
  return(0);
}


/**********************************************************************
 * SB_put_bug_cell
 * Put a bug in a cell (or remove a bug from a cell)
 **********************************************************************/

void SB_put_bug_cell(char **bugs, char ch, int cell){
    *((*bugs)+cell)=ch;
    SBtt_putchar_cell(ch, cell);
}


/**********************************************************************
 * SB_put_bug_rc
 * Put a bug in a row, column co-ordinate
 **********************************************************************/

void SB_put_bug_rc(char **bugs, char ch, int r, int c){
    extern int SLtt_Screen_Cols;
    
    *((*bugs)+c+(r*SLtt_Screen_Cols))=ch;
    SBtt_putchar_rc(ch, r, c);
}


/**********************************************************************
 * SBtt_putchar_rc
 * Put a character in a row an column co-cordinate
 **********************************************************************/

void SBtt_putchar_rc(char ch, int r, int c){
    SLtt_goto_rc(r, c);
    SLtt_putchar(ch);
 }


/**********************************************************************
 * SBtt_putchar_cell
 * Put a character in a row an cell location
 **********************************************************************/

void SBtt_putchar_cell(char ch, int cell){
    extern int SLtt_Screen_Cols;

    SBtt_putchar_rc(ch, cell/SLtt_Screen_Cols, cell%SLtt_Screen_Cols);
 }


/**********************************************************************
 * SB_get_key
 * Non blocking get key implimented using SLang_input_pending
 * Argument as for SLang_input_pending
 * Return: 0 if no input ready
 *         input as per SLang_getkey, otherwise
 **********************************************************************/

int SB_get_key(int tsecs){
  if(SLang_input_pending(tsecs) == 1){
    return(SLang_getkey());
  }

  return(0);
}


/**********************************************************************
 * SB_kill_bug_mouse
 * Kill a bug according to a mouse click
 * Return: 0 if no bugs killed
 *         number of bugs killed (1) otherwise
 **********************************************************************/

int SB_kill_bug_mouse(
  char **bugs, 
  int x, 
  int y, 
  int b, 
  int range,
  int cells
){
  extern int SLtt_Screen_Cols;
  extern int SLtt_Screen_Rows;
  
  if(x<0 || y<0 || x>=SLtt_Screen_Cols || y>(SLtt_Screen_Rows-2)){
    return(0);
  }

  if(*((*bugs)+x+(y*SLtt_Screen_Cols))!=BUG){
    if(range!=0){
      if(SB_kill_bug_mouse(bugs, x-1, y  , b, range-1, cells)){ return(1); }
      if(SB_kill_bug_mouse(bugs, x-1, y+1, b, range-1, cells)){ return(1); }
      if(SB_kill_bug_mouse(bugs, x  , y+1, b, range-1, cells)){ return(1); }
      if(SB_kill_bug_mouse(bugs, x+1, y+1, b, range-1, cells)){ return(1); }
      if(SB_kill_bug_mouse(bugs, x+1, y  , b, range-1, cells)){ return(1); }
      if(SB_kill_bug_mouse(bugs, x+1, y-1, b, range-1, cells)){ return(1); }
      if(SB_kill_bug_mouse(bugs, x  , y-1, b, range-1, cells)){ return(1); }
      if(SB_kill_bug_mouse(bugs, x-1, y-1, b, range-1, cells)){ return(1); }
    }
    return(0);
  }

  SB_put_bug_rc(bugs, DEADBUG, y, x);
  return(1);
}
    
